class HttpError(RuntimeError):
    code = ''
    message = ''
    headers = {}

    def __init__(self, message=None):
        self.message = message or self.message
        super(HttpError, self).__init__(self.message)


class HttpUnauthorized(HttpError):
    code = '401 Unauthorized'
    def __init__(self, auth_scheme=None):
        if not auth_scheme:
            auth_scheme = 'Mushroom'
        self.headers = [
            ('WWW-Authenticate', auth_scheme)
        ]
        super(HttpUnauthorized, self).__init__()


class HttpNotFound(HttpError):
    code = '404 Not Found'


class HttpMethodNotAllowed(HttpError):
    code = '405 Method Not Allowed'
    def __init__(self, allowed_methods):
        self.headers = [
            ('Allow', ', '.join(method.upper() for method in allowed_methods))
        ]
        super(HttpMethodNotAllowed, self).__init__()


class HttpInternalServerError(HttpError):
    code = '500 Internal Server Error'


class HttpNotImplemented(HttpError):
    code = '501 Not Implemented'



from itertools import count
import random
import string
from time import time

import gevent
from gevent.event import Event
try:
    import simplejson as json
except ImportError:
    import json

from mushroom.http import JsonResponse
from mushroom.rpc import Engine as RpcEngine
from mushroom.rpc import Message
from mushroom.rpc import Heartbeat
from mushroom.rpc import Request
from mushroom.rpc import Notification
from mushroom.transport import UnreliableTransport


POLL_TIMEOUT = 40
SESSION_CLEANUP_INTERVAL = 30
SESSION_MAX_AGE = 60


def session_id_generator():
    alpha_numerics = string.ascii_letters + string.digits
    while True:
        '''A UUID has 16 bytes and therefore 256^16 possibilities. For the
        shortest possible id which consists only of URL safe characters we
        pick only alpha numeric characters (2*26+10) and make it 22
        characters long. This is at least as good as a hex encoded UUID
        which would require 32 characters.'''
        yield ''.join(random.choice(alpha_numerics) for _ in xrange(22))


class Session(RpcEngine):

    def __init__(self, id, transport, rpc_handler):
        self.id = id
        super(Session, self).__init__(transport, rpc_handler)

    def get_handshake_data(self, request):
        return self.transport.get_handshake_data(request)

    def handle_disconnect(self):
        # FIXME make sure the session is removed from the session list
        pass



class HttpTransport(UnreliableTransport):

    def get_url(self, request, protocol):
        host = request.environ['HTTP_HOST']
        return '%s://%s/%s/' % (protocol, host, self.rpc_engine.id)


class PollTransport(HttpTransport):

    timeout = 40

    def __init__(self):
        super(PollTransport, self).__init__()
        self.messages_ready = Event()

    def get_handshake_data(self, request):
        protocol = 'http' # XXX autodetect
        return {
            'transport': 'poll',
            'url': self.get_url(request, protocol)
        }

    def handle_http_request(self, request, session):
        # Only allow POST requests for polling
        if request.method != 'POST':
            raise HttpMethodNotAllowed(['POST'])
        assert isinstance(request.data, list)
        heartbeat = None
        for message_data in request.data:
            message = Message.from_list(message_data)
            message.session = session
            self.handle_message(message)
            if isinstance(message, Heartbeat):
                heartbeat = message
        if heartbeat:
            self.handle_connect()
            self.messages_ready.wait(self.timeout)
            self.messages_ready.clear()
            self.handle_disconnect(reconnect=True)
            return JsonResponse([
                message.to_list()
                for message in self.messages
            ])
        else:
            return JsonResponse(None)

    def real_send(self, message):
        self.messages_ready.set()


class WebSocketTransport(HttpTransport):

    def __init__(self):
        super(WebSocketTransport, self).__init__()
        self.ws = None

    def get_handshake_data(self, request):
        protocol = 'ws' # XXX autodetect
        return {
            'transport': 'ws',
            'url': self.get_url(request, protocol)
        }

    def handle_http_request(self, request, session):
        assert not self.ws
        self.ws = request.environ['wsgi.websocket']
        self.handle_connect()
        # Process incoming messages
        try:
            while True:
                frame = self.ws.receive()
                if frame is None:
                    # Disconnect
                    return
                message_data = json.loads(frame)
                message = Message.from_list(message_data)
                message.session = session
                self.handle_message(message)
        finally:
            self.handle_disconnect()

    def handle_disconnect(self):
        self.ws = None
        super(WebSocketTransport, self).handle_disconnect()

    def real_send(self, message):
        message_data = message.to_list()
        frame = json.dumps(message_data)
        # FIXME this can fail if the websocket goes down
        self.ws.send(frame)


class SessionList(object):

    def __init__(self):
        self.sessions = {}

    def add(self, session):
        if session.id in self.sessions:
            raise KeyError('Duplicate session id %r' % session.id)
        self.sessions[session.id] = session

    def remove(self, session):
        del self.sessions[session.id]

    def notify(self, method, data=None):
        for session in self.sessions.itervalues():
            session.notify(method, data)

    def __getitem__(self, sid):
        return self.sessions[sid]



from itertools import count
import logging

import gevent
from gevent.event import Event


logger = logging.getLogger('mushroom.rpc')


class RpcError(RuntimeError):
    pass


class MethodNotFound(RpcError):
    pass


class RequestTimeout(RpcError):
    pass


class RequestException(RpcError):

    def __init__(self, data):
        super(RequestException, self).__init__()
        self.data = data


class MethodDispatcher(object):

    def __init__(self, obj, prefix='rpc_'):
        self.obj = obj
        self.prefix = prefix

    def __call__(self, request):
        method_name = self.prefix + request.method
        try:
            method = getattr(self.obj, method_name)
        except AttributeError:
            raise MethodNotFound(method_name)
        return method(request)


def dummy_rpc_handler(request):
    '''
    Dummy RPC handler that raises a MethodNotFound exception for
    all calls. This is useful for applications that do not need do
    receive any data from the client but only publish data.
    '''
    raise MethodNotFound(request.method)


class Engine(object):
    '''Transport neutral message factory and mapper between requests and
    responses.'''

    def __init__(self, transport, rpc_handler):
        # Transport for sending and receiving messages
        self.transport = transport
        # Bind transport to the engine
        assert transport.rpc_engine is None, \
                'transport already bound to another rpc engine'
        transport.rpc_engine = self
        # Handler for inbound requests and notifications
        self.rpc_handler = rpc_handler
        # Generator for outbount message ids
        self.message_id_generator = count()
        # Dictionary for mapping outbound requests to inbound responses
        self.requests = {}

    def next_message_id(self):
        '''Generate the next message id for outbound messages.'''
        return self.message_id_generator.next()

    def notify(self, method, data=None, **kwargs):
        '''Send a notification.'''
        message = Notification(method, data, message_id=self.next_message_id())
        self.send(message, **kwargs)

    def request(self, method, data=None, timeout=None, **kwargs):
        '''Send a request and wait for the response or timeout.'''
        request = Request(method, data, message_id=self.next_message_id())
        self.requests[request.message_id] = request
        self.send(request, **kwargs)
        response = request.get_response(timeout)
        if isinstance(response, Response):
            return response.data
        elif isinstance(response, Error):
            raise RequestException(response.data)

    def send(self, message, **kwargs):
        '''Hand message over to the transport.'''
        self.transport.send(message, **kwargs)

    def handle_message(self, message):
        '''Handle message received from the transport.'''
        if isinstance(message, Notification):
            # Spawn worker to process the notification. The response of
            # the worker is ignored.
            def worker():
                try:
                    self.rpc_handler(message)
                except MethodNotFound, e:
                    logger.warning('MethodNotFound: %s' % e.message)
                except RpcError, e:
                    logger.debug(e, exc_info=True)
                except Exception, e:
                    logger.exception(e)
            gevent.spawn(worker)
        elif isinstance(message, Request):
            # Spawn worker which waits for the response of the rpc handler
            # and sends the response message.
            def worker():
                try:
                    response_data = self.rpc_handler(message)
                except MethodNotFound, e:
                    logger.warning('MethodNotFound: %s' % e.message)
                    self.send(Error(message, e.message,
                        message_id=self.next_message_id()))
                except RpcError, e:
                    logger.debug(e, exc_info=True)
                    self.send(Error(message, e.message,
                        message_id=self.next_message_id()))
                except Exception, e:
                    logger.exception(e)
                    self.send(Error(message, 'Internal server error',
                        message_id=self.next_message_id()))
                else:
                    self.send(Response(message, response_data,
                        message_id=self.next_message_id()))
            gevent.spawn(worker)
        elif isinstance(message, (Response, Error)):
            # Find request according to response or error.
            try:
                request = self.requests.pop(message.request)
            except KeyError:
                logger.error('Response for unknown request message id: %r' %
                        message.request)
                return
            request.response = message
        else:
            raise RuntimeError('Unsupported message type: %s' % type(message))


class Message(object):

    @staticmethod
    def from_list(l):
        if not isinstance(l, (list, tuple)):
            raise ValueError('Message is not encoded as list or tuple')
        try:
            message_class = MESSAGE_CLASS_BY_CODE[l[0]]
        except KeyError:
            raise ValueError('Unsupported message code: %r' % l[0])
        message = message_class.from_list(l)
        return message


class Heartbeat(Message):
    code = 0

    def __init__(self, last_message_id):
        self.last_message_id = last_message_id

    @staticmethod
    def from_list(l):
        self = Heartbeat(l[1])
        return self


class Notification(Message):
    code = 1
    message_id = None

    def __init__(self, method, data=None, message_id=None):
        self.method = method
        self.data = data
        self.message_id = message_id

    @staticmethod
    def from_list(l):
        self = Notification(l[2], l[3], message_id=l[1])
        return self

    def to_list(self):
        return [self.code, self.message_id, self.method, self.data]


class Request(Message):
    code = 2
    message_id = None

    def __init__(self, method, data=None, message_id=None):
        self.method = method
        self.data = data
        self.message_id = message_id
        self._response = None
        self.complete = Event()

    def to_list(self):
        return [self.code, self.message_id, self.method, self.data]

    @staticmethod
    def from_list(l):
        self = Request(l[2], l[3], message_id=l[1])
        return self

    @property
    def response(self):
        return self._response

    @response.setter
    def response(self, response):
        assert not self._response
        self._response = response
        self.complete.set()

    def get_response(self, timeout=None):
        if self.complete.wait(timeout):
            return self.response
        else:
            raise RequestTimeout


class Response(Message):
    code = 3
    message_id = None

    def __init__(self, request, data, message_id=None):
        self.request = request
        self.data = data
        self.message_id = message_id

    @staticmethod
    def from_list(l):
        self = Response(l[2], l[3], message_id=l[1])
        return self

    def to_list(self):
        return [self.code, self.message_id, self.request.message_id, self.data]


class Error(Message):
    code = 4
    message_id = None

    def __init__(self, request, data, message_id=None):
        self.request = request
        self.data = data
        self.message_id = message_id

    @staticmethod
    def from_list(l):
        self = Error(l[2], l[3], message_id=l[1])
        return self

    def to_list(self):
        return [self.code, self.message_id, self.request.message_id, self.data]


class Disconnect(Message):
    code = -1

    @staticmethod
    def from_list(l):
        self = Disconnect()
        return self


MESSAGE_CLASS_BY_CODE = {
    Heartbeat.code: Heartbeat,
    Notification.code: Notification,
    Request.code: Request,
    Response.code: Response,
    Error.code: Error,
    Disconnect.code: Disconnect
}



from gevent.event import Event

from mushroom.rpc import Heartbeat
from mushroom.rpc import Disconnect
from mushroom.rpc import Message


class UnreliableTransport(object):
    '''
    Base class for unreliable transports. Unreliable means that
    the underlying protocol does not guarantee message delivery
    by itself. This is typically the case for any protocol that
    does raw communication without a layer that guarantees message
    delivery by itself.
    '''

    def __init__(self):
        '''
        Keyword arguments:
        rpc_engine  -- Handler for RPC messages
        '''
        # Outbound messages not yet acknowledged by the remote side
        self.messages = []
        # Last received message id. This is used to drop messages
        # already received from the remote side.
        self.last_message_id = None
        # Handler for RPC messages
        self.rpc_engine = None
        # Is this transport currently connected?
        self.connected = False
        # Is this transport currently connecting?
        self.connecting = False

    def handle_message(self, message):
        if hasattr(message, 'message_id'):
            if message.message_id <= self.last_message_id:
                # Ignore messages that we have already received
                return
        if isinstance(message, Heartbeat):
            self.handle_heartbeat(message)
        elif isinstance(message, Disconnect):
            self.handle_disconnect()
        else:
            # All messages except for Heartbeat and Disconnect
            # are handled by the RPC engine.
            self.rpc_engine.handle_message(message)

    def handle_heartbeat(self, heartbeat):
        last_message_id = heartbeat.last_message_id
        index = 0
        for message in self.messages:
            if message.message_id <= last_message_id:
                index += 1
            else:
                break
        if index > 0:
            del self.messages[:index]

    def handle_connect(self):
        # The client managed to connect. Try resending messages which
        # have not been acknowledged, yet.
        self.connecting = False
        self.connected = True
        for message in self.messages:
            self.real_send(message)

    def handle_disconnect(self, reconnect=False):
        self.connecting = reconnect
        self.connected = False

    def connect(self):
        '''
        Start transport and connect to the remote side. This method
        is transport specific and must be overwriten.
        '''
        raise NotImplementedError

    def disconnect(self):
        '''
        Disconnect from the remote side and stop transport. This method
        is transport specific and must be overwritten.
        '''
        raise NotImplementedError

    def send(self, message):
        self.messages.append(message)
        if self.connected:
            self.real_send(message)

    def real_send(self, message):
        '''
        Perform the actual send operation. This method is only called
        internally and should not be called from application code.
        This method is transport specific and must be overwritten.
        '''
        raise NotImplementedError



from gevent.pywsgi import WSGIServer
from geventwebsocket.handler import WebSocketHandler

from mushroom.application import Application


class Server(WSGIServer):

    def __init__(self, listener, rpc_handler=None,
            auth_handler=None, disconnect_handler=None):
        super(Server, self).__init__(listener,
                Application(rpc_handler, auth_handler, disconnect_handler),
                handler_class=WebSocketHandler)

    @property
    def sessions(self):
        return self.application.sessions



try:
    import simplejson as json
except:
    import json


class HttpRequest(object):

    def __init__(self, environ):
        # FIXME add validation
        self.environ = environ
        self.method = environ['REQUEST_METHOD']
        self.raw_data = environ['wsgi.input'].read()
        if self.raw_data:
            self.data = json.loads(self.raw_data)
        else:
            self.data = None
        # Convert path to list of atoms
        self.path = filter(bool, environ['PATH_INFO'].split('/'))


class HttpResponse(object):

    def __init__(self, code='200 OK', content='', extra_headers=None):
        self.code = code
        self.content = content
        self.headers = {
            'Access-Control-Allow-Origin': '*',
            'Content-Type': 'text/plain',
        }
        self.headers.update(extra_headers or {})


class JsonResponse(HttpResponse):

    def __init__(self, data):
        super(JsonResponse, self).__init__(
            content=json.dumps(data),
            extra_headers={
                'Content-Type': 'application/json',
            }
        )
